from django.contrib.postgres.fields import JSONField
from django.db import models


class CustomFieldMixin:
    pass


class RestrictedCharField(models.CharField):
    def __init__(self, *args, **kwargs):
        kwargs["max_length"] = 20
        super().__init__(*args, **kwargs)


class NotNullRestrictedCharField(RestrictedCharField):
    def __init__(self, *args, **kwargs):
        kwargs["null"] = False
        kwargs["blank"] = False
        super().__init__(*args, **kwargs)


class ListField(CustomFieldMixin, JSONField):
    def __init__(self, *args, **kwargs):
        kwargs["default"] = list
        super().__init__(*args, **kwargs)



import pytest
from testapp.models import Configuration, Listing, Preference, Record, Searchable

from djantic import ModelSchema


@pytest.mark.django_db
def test_unhandled_field_type():
    class SearchableSchema(ModelSchema):
        class Config:
            model = Searchable

    assert SearchableSchema.schema() == {
        "title": "SearchableSchema",
        "description": "Searchable(id, title, search_vector)",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "title": {
                "title": "Title",
                "description": "title",
                "maxLength": 255,
                "type": "string",
            },
            "search_vector": {
                "title": "Search Vector",
                "description": "search_vector",
                "type": "string",
            },
        },
        "required": ["title"],
    }

    searchable = Searchable.objects.create(title="My content")
    assert SearchableSchema.from_django(searchable).dict() == {
        "id": 1,
        "title": "My content",
        "search_vector": None,
    }


@pytest.mark.django_db
def test_custom_field():
    """
    Test a model using custom field subclasses.
    """

    class RecordSchema(ModelSchema):
        class Config:
            model = Record
            include = ["id", "title", "items"]

    assert RecordSchema.schema() == {
        "title": "RecordSchema",
        "description": "A generic record model.",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "title": {
                "title": "Title",
                "description": "title",
                "maxLength": 20,
                "type": "string",
            },
            "items": {
                "title": "Items",
                "description": "items",
                "anyOf": [
                    {"type": "string", "format": "json-string"},
                    {"type": "object"},
                    {"type": "array", "items": {}},
                ],
            },
        },
        "required": ["title"],
    }


@pytest.mark.django_db
def test_postgres_json_field():
    """
    Test generating a schema for multiple Postgres JSON fields.
    """

    class ConfigurationSchema(ModelSchema):
        class Config:
            model = Configuration
            include = ["permissions", "changelog", "metadata"]

    assert ConfigurationSchema.schema() == {
        "title": "ConfigurationSchema",
        "description": "A configuration container.",
        "type": "object",
        "properties": {
            "permissions": {
                "title": "Permissions",
                "description": "permissions",
                "anyOf": [
                    {"type": "string", "format": "json-string"},
                    {"type": "object"},
                    {"type": "array", "items": {}},
                ],
            },
            "changelog": {
                "title": "Changelog",
                "description": "changelog",
                "anyOf": [
                    {"type": "string", "format": "json-string"},
                    {"type": "object"},
                    {"type": "array", "items": {}},
                ],
            },
            "metadata": {
                "title": "Metadata",
                "description": "metadata",
                "anyOf": [
                    {"type": "string", "format": "json-string"},
                    {"type": "object"},
                    {"type": "array", "items": {}},
                ],
            },
        },
    }


@pytest.mark.django_db
def test_lazy_choice_field():
    """
    Test generating a dynamic enum choice field.
    """

    class RecordSchema(ModelSchema):
        class Config:
            model = Record
            include = ["record_type", "record_status"]

    assert RecordSchema.schema() == {
        "title": "RecordSchema",
        "description": "A generic record model.",
        "type": "object",
        "properties": {
            "record_type": {
                "title": "Record Type",
                "description": "record_type",
                "default": "NEW",
                "allOf": [{"$ref": "#/definitions/RecordSchemaRecordTypeEnum"}],
            },
            "record_status": {
                "title": "Record Status",
                "description": "record_status",
                "default": 0,
                "allOf": [{"$ref": "#/definitions/RecordSchemaRecordStatusEnum"}],
            },
        },
        "definitions": {
            "RecordSchemaRecordTypeEnum": {
                "title": "RecordSchemaRecordTypeEnum",
                "description": "An enumeration.",
                "enum": ["NEW", "OLD"],
            },
            "RecordSchemaRecordStatusEnum": {
                "title": "RecordSchemaRecordStatusEnum",
                "description": "An enumeration.",
                "enum": [0, 1, 2],
            },
        },
    }


@pytest.mark.django_db
def test_enum_choices():
    class PreferenceSchema(ModelSchema):
        class Config:
            model = Preference
            use_enum_values = True

    assert PreferenceSchema.schema() == {
        "title": "PreferenceSchema",
        "description": "Preference(id, name, preferred_food, preferred_group, preferred_sport, preferred_musician)",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "name": {
                "title": "Name",
                "description": "name",
                "maxLength": 128,
                "type": "string",
            },
            "preferred_food": {
                "title": "Preferred Food",
                "description": "preferred_food",
                "default": "ba",
                "allOf": [{"$ref": "#/definitions/PreferenceSchemaPreferredFoodEnum"}],
            },
            "preferred_group": {
                "title": "Preferred Group",
                "description": "preferred_group",
                "default": 1,
                "allOf": [{"$ref": "#/definitions/PreferenceSchemaPreferredGroupEnum"}],
            },
            "preferred_sport": {
                "title": "Preferred Sport",
                "description": "preferred_sport",
                "allOf": [{"$ref": "#/definitions/PreferenceSchemaPreferredSportEnum"}],
            },
            "preferred_musician": {
                "allOf": [{"$ref": "#/definitions/PreferenceSchemaPreferredMusicianEnum"}],
                'default': '',
                "description": "preferred_musician",
                "title": "Preferred Musician"
            },
        },
        "required": ["name"],
        "definitions": {
            "PreferenceSchemaPreferredFoodEnum": {
                "title": "PreferenceSchemaPreferredFoodEnum",
                "description": "An enumeration.",
                "enum": ["ba", "ap"],
            },
            "PreferenceSchemaPreferredGroupEnum": {
                "title": "PreferenceSchemaPreferredGroupEnum",
                "description": "An enumeration.",
                "enum": [1, 2],
            },
            "PreferenceSchemaPreferredSportEnum": {
                "title": "PreferenceSchemaPreferredSportEnum",
                "description": "An enumeration.",
                "enum": ["football", "basketball", ""],
            },
            "PreferenceSchemaPreferredMusicianEnum": {
                "title": "PreferenceSchemaPreferredMusicianEnum",
                "description": "An enumeration.",
                "enum": ["tom_jobim", "sinatra", ""],
            }
        },
    }

    preference = Preference.objects.create(name="Jordan", preferred_sport="", preferred_musician=None)
    assert PreferenceSchema.from_django(preference).dict() == {
        "id": 1,
        "name": "Jordan",
        "preferred_food": "ba",
        "preferred_group": 1,
        "preferred_sport": "",
        'preferred_musician': None
    }


@pytest.mark.django_db
def test_enum_choices_generates_unique_enums():
    class PreferenceSchema(ModelSchema):
        class Config:
            model = Preference
            use_enum_values = True

    class PreferenceSchema2(ModelSchema):
        class Config:
            model = Preference
            use_enum_values = True

    assert str(PreferenceSchema2.__fields__["preferred_food"].type_) != str(
        PreferenceSchema.__fields__["preferred_food"].type_
    )


@pytest.mark.django_db
def test_listing():
    class ListingSchema(ModelSchema):
        class Config:
            model = Listing
            use_enum_values = True

    assert ListingSchema.schema() == {
        "title": "ListingSchema",
        "description": "Listing(id, items)",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "items": {
                "description": "items",
                "items": {},
                "title": "Items",
                "type": "array",
            },
        },
        "required": ["items"],
    }

    preference = Listing(items=["a", "b"])
    assert ListingSchema.from_django(preference).dict() == {
        "id": None,
        "items": ["a", "b"],
    }




from decimal import Decimal
from typing import List, Optional

import pytest
from pydantic import validator
from testapp.order import Order, OrderItem, OrderItemDetail, OrderUser, OrderUserFactory, OrderUserProfile

from djantic import ModelSchema


@pytest.mark.django_db
def test_multiple_level_relations():
    class OrderItemDetailSchema(ModelSchema):
        class Config:
            model = OrderItemDetail

    class OrderItemSchema(ModelSchema):
        details: List[OrderItemDetailSchema]

        class Config:
            model = OrderItem

    class OrderSchema(ModelSchema):
        items: List[OrderItemSchema]

        class Config:
            model = Order

    class OrderUserProfileSchema(ModelSchema):

        class Config:
            model = OrderUserProfile

    class OrderUserSchema(ModelSchema):
        orders: List[OrderSchema]
        profile: OrderUserProfileSchema
        user_cache: Optional[dict]

        class Config:
            model = OrderUser
            include = ('id',
                       'first_name',
                       'last_name',
                       'email',
                       'profile',
                       'orders',
                       'user_cache')

        @validator('user_cache', pre=True, always=True)
        def get_user_cache(cls, _):
            return {
                'has_order': True
            }

    user = OrderUserFactory.create()

    assert OrderUserSchema.from_django(user).dict() == {
        'id': 1,
        'first_name': '',
        'last_name': None,
        'email': '',
        'user_cache': {'has_order': True},
        'profile': {
            'id': 1,
            'address': '',
            'user': 1
        },
        'orders': [
            {
                'id': 1,
                'total_price': Decimal('0.00000'),
                'shipping_address': '',
                'user': 1,
                'items': [
                    {
                        'id': 1,
                        'name': '',
                        'price': Decimal('0.00000'),
                        'quantity': 0,
                        'order': 1,
                        'details': [
                            {
                                'id': 1,
                                'name': '',
                                'value': 0,
                                'quantity': 0,
                                'order_item': 1
                            },
                            {
                                'id': 2,
                                'name': '',
                                'value': 0,
                                'quantity': 0,
                                'order_item': 1
                            }
                        ]
                    },
                    {
                        'details': [
                            {
                                'id': 3,
                                'name': '',
                                'value': 0,
                                'quantity': 0,
                                'order_item': 2
                            },
                            {
                                'id': 4,
                                'name': '',
                                'value': 0,
                                'quantity': 0,
                                'order_item': 2
                            }
                        ],
                        'id': 2,
                        'name': '',
                        'price': Decimal('0.00000'),
                        'quantity': 0,
                        'order': 1
                    }
                ],

            },
            {
                'id': 2,
                'total_price': Decimal('0.00000'),
                'shipping_address': '',
                'user': 1,
                'items': [
                    {
                        'id': 3,
                        'name': '',
                        'price': Decimal('0.00000'),
                        'quantity': 0,
                        'order': 2,
                        'details': [
                            {
                                'id': 5,
                                'name': '',
                                'value': 0,
                                'quantity': 0,
                                'order_item': 3},
                            {
                                'id': 6,
                                'name': '',
                                'value': 0,
                                'quantity': 0,
                                'order_item': 3}
                        ]
                    },
                    {
                        'id': 4,
                        'name': '',
                        'price': Decimal('0.00000'),
                        'quantity': 0,
                        'order': 2,
                        'details': [
                            {
                                'id': 7,
                                'name': '',
                                'value': 0,
                                'quantity': 0,
                                'order_item': 4},
                            {
                                'id': 8,
                                'name': '',
                                'value': 0,
                                'quantity': 0,
                                'order_item': 4}]
                    }
                ]
            }
        ]
    }

    assert OrderUserSchema.schema() == {
        "title": "OrderUserSchema",
        "description": "OrderUser(id, first_name, last_name, email)",
        "type": "object",
        "properties": {
            "profile": {
                "$ref": "#/definitions/OrderUserProfileSchema"
            },
            "orders": {
                "title": "Orders",
                "type": "array",
                "items": {
                    "$ref": "#/definitions/OrderSchema"
                }
            },
            "id": {
                "title": "Id",
                "description": "id",
                "type": "integer"
            },
            "first_name": {
                "title": "First Name",
                "description": "first_name",
                "maxLength": 50,
                "type": "string"
            },
            "last_name": {
                "title": "Last Name",
                "description": "last_name",
                "maxLength": 50,
                "type": "string"
            },
            "email": {
                "title": "Email",
                "description": "email",
                "maxLength": 254,
                "type": "string"
            },
            "user_cache": {
                "title": "User Cache",
                "type": "object"
            }
        },
        "required": [
            "profile",
            "orders",
            "first_name",
            "email"
        ],
        "definitions": {
            "OrderUserProfileSchema": {
                "title": "OrderUserProfileSchema",
                "description": "OrderUserProfile(id, address, user)",
                "type": "object",
                "properties": {
                    "id": {
                        "title": "Id",
                        "description": "id",
                        "type": "integer"
                    },
                    "address": {
                        "title": "Address",
                        "description": "address",
                        "maxLength": 255,
                        "type": "string"
                    },
                    "user": {
                        "title": "User",
                        "description": "id",
                        "type": "integer"
                    }
                },
                "required": [
                    "address",
                    "user"
                ]
            },
            "OrderItemDetailSchema": {
                "title": "OrderItemDetailSchema",
                "description": "OrderItemDetail(id, name, value, quantity, order_item)",
                "type": "object",
                "properties": {
                    "id": {
                        "title": "Id",
                        "description": "id",
                        "type": "integer"
                    },
                    "name": {
                        "title": "Name",
                        "description": "name",
                        "maxLength": 30,
                        "type": "string"
                    },
                    "value": {
                        "title": "Value",
                        "description": "value",
                        "default": 0,
                        "type": "integer"
                    },
                    "quantity": {
                        "title": "Quantity",
                        "description": "quantity",
                        "default": 0,
                        "type": "integer"
                    },
                    "order_item": {
                        "title": "Order Item",
                        "description": "id",
                        "type": "integer"
                    }
                },
                "required": [
                    "name",
                    "order_item"
                ]
            },
            "OrderItemSchema": {
                "title": "OrderItemSchema",
                "description": "OrderItem(id, name, price, quantity, order)",
                "type": "object",
                "properties": {
                    "details": {
                        "title": "Details",
                        "type": "array",
                        "items": {
                            "$ref": "#/definitions/OrderItemDetailSchema"
                        }
                    },
                    "id": {
                        "title": "Id",
                        "description": "id",
                        "type": "integer"
                    },
                    "name": {
                        "title": "Name",
                        "description": "name",
                        "maxLength": 30,
                        "type": "string"
                    },
                    "price": {
                        "title": "Price",
                        "description": "price",
                        "default": 0,
                        "type": "number"
                    },
                    "quantity": {
                        "title": "Quantity",
                        "description": "quantity",
                        "default": 0,
                        "type": "integer"
                    },
                    "order": {
                        "title": "Order",
                        "description": "id",
                        "type": "integer"
                    }
                },
                "required": [
                    "details",
                    "name",
                    "order"
                ]
            },
            "OrderSchema": {
                "title": "OrderSchema",
                "description": "Order(id, total_price, shipping_address, user)",
                "type": "object",
                "properties": {
                    "items": {
                        "title": "Items",
                        "type": "array",
                        "items": {
                            "$ref": "#/definitions/OrderItemSchema"
                        }
                    },
                    "id": {
                        "title": "Id",
                        "description": "id",
                        "type": "integer"
                    },
                    "total_price": {
                        "title": "Total Price",
                        "description": "total_price",
                        "default": 0,
                        "type": "number"
                    },
                    "shipping_address": {
                        "title": "Shipping Address",
                        "description": "shipping_address",
                        "maxLength": 255,
                        "type": "string"
                    },
                    "user": {
                        "title": "User",
                        "description": "id",
                        "type": "integer"
                    }
                },
                "required": [
                    "items",
                    "shipping_address",
                    "user"
                ]
            }
        }
    }



import pytest
from pydantic import ConfigError
from testapp.models import User

from djantic import ModelSchema


@pytest.mark.django_db
def test_config_errors():
    """
    Test the model config error exceptions.
    """

    with pytest.raises(
        ConfigError, match="(Is `Config` class defined?)"
    ):

        class InvalidModelErrorSchema(ModelSchema):
            pass

    with pytest.raises(
        ConfigError, match="(Is `Config.model` a valid Django model class?)"
    ):

        class InvalidModelErrorSchema(ModelSchema):
            class Config:
                model = "Ok"

    with pytest.raises(
        ConfigError,
        match="Only one of 'include' or 'exclude' should be set in configuration.",
    ):

        class IncludeExcludeErrorSchema(ModelSchema):
            class Config:
                model = User
                include = ["id"]
                exclude = ["first_name"]


@pytest.mark.django_db
def test_get_field_names():
    """
    Test retrieving the field names for a model.
    """

    class UserSchema(ModelSchema):
        class Config:
            model = User
            include = ["id"]

    assert UserSchema.get_field_names() == ["id"]

    class UserSchema(ModelSchema):
        class Config:
            model = User
            exclude = ["id"]

    assert UserSchema.get_field_names() == [
        "profile",
        "first_name",
        "last_name",
        "email",
        "created_at",
        "updated_at",
    ]



import datetime
from typing import Dict, List, Optional

import pytest
from pydantic import Field
from testapp.models import (
    Article,
    Bookmark,
    Case,
    Expert,
    Item,
    Message,
    Profile,
    Publication,
    Tagged,
    Thread,
    User
)

from djantic import ModelSchema


@pytest.mark.django_db
def test_m2m():
    """
    Test forward m2m relationships.
    """

    class ArticleSchema(ModelSchema):
        class Config:
            model = Article

    assert ArticleSchema.schema() == {
        "title": "ArticleSchema",
        "description": "A news article.",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "headline": {
                "title": "Headline",
                "description": "headline",
                "maxLength": 100,
                "type": "string",
            },
            "pub_date": {
                "title": "Pub Date",
                "description": "pub_date",
                "type": "string",
                "format": "date",
            },
            "publications": {
                "title": "Publications",
                "description": "id",
                "type": "array",
                "items": {
                    "type": "object",
                    "additionalProperties": {"type": "integer"},
                },
            },
        },
        "required": ["headline", "pub_date", "publications"],
    }

    class PublicationSchema(ModelSchema):
        class Config:
            model = Publication

    class ArticleWithPublicationListSchema(ModelSchema):
        publications: List[PublicationSchema]

        class Config:
            model = Article

    assert ArticleWithPublicationListSchema.schema() == {
        "title": "ArticleWithPublicationListSchema",
        "description": "A news article.",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "headline": {
                "title": "Headline",
                "description": "headline",
                "maxLength": 100,
                "type": "string",
            },
            "pub_date": {
                "title": "Pub Date",
                "description": "pub_date",
                "type": "string",
                "format": "date",
            },
            "publications": {
                "title": "Publications",
                "type": "array",
                "items": {"$ref": "#/definitions/PublicationSchema"},
            },
        },
        "required": ["headline", "pub_date", "publications"],
        "definitions": {
            "PublicationSchema": {
                "title": "PublicationSchema",
                "description": "A news publication.",
                "type": "object",
                "properties": {
                    "article_set": {
                        "title": "Article Set",
                        "description": "id",
                        "type": "array",
                        "items": {
                            "type": "object",
                            "additionalProperties": {"type": "integer"},
                        },
                    },
                    "id": {"title": "Id", "description": "id", "type": "integer"},
                    "title": {
                        "title": "Title",
                        "description": "title",
                        "maxLength": 30,
                        "type": "string",
                    },
                },
                "required": ["title"],
            }
        },
    }

    article = Article.objects.create(
        headline="My Headline", pub_date=datetime.date(2021, 3, 20)
    )
    publication = Publication.objects.create(title="My Publication")
    article.publications.add(publication)

    schema = ArticleWithPublicationListSchema.from_django(article)
    assert schema.dict() == {
        "id": 1,
        "headline": "My Headline",
        "pub_date": datetime.date(2021, 3, 20),
        "publications": [{"article_set": [{'id': 1}], "id": 1, "title": "My Publication"}],
    }


@pytest.mark.django_db
def test_foreign_key():
    """
    Test forward foreign-key relationships.
    """

    class ThreadSchema(ModelSchema):
        class Config:
            model = Thread

    class MessageSchema(ModelSchema):
        class Config:
            model = Message

    assert MessageSchema.schema() == {
        "title": "MessageSchema",
        "description": "A message posted in a thread.",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "content": {"title": "Content", "description": "content", "type": "string"},
            "created_at": {
                "title": "Created At",
                "description": "created_at",
                "type": "string",
                "format": "date-time",
            },
            "thread": {"title": "Thread", "description": "id", "type": "integer"},
        },
        "required": ["content", "created_at", "thread"],
    }

    class MessageWithThreadSchema(ModelSchema):

        thread: ThreadSchema

        class Config:
            model = Message

    assert MessageWithThreadSchema.schema() == {
        "title": "MessageWithThreadSchema",
        "description": "A message posted in a thread.",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "content": {"title": "Content", "description": "content", "type": "string"},
            "created_at": {
                "title": "Created At",
                "description": "created_at",
                "type": "string",
                "format": "date-time",
            },
            "thread": {"$ref": "#/definitions/ThreadSchema"},
        },
        "required": ["content", "created_at", "thread"],
        "definitions": {
            "ThreadSchema": {
                "title": "ThreadSchema",
                "description": "A thread of messages.",
                "type": "object",
                "properties": {
                    "messages": {
                        "title": "Messages",
                        "description": "id",
                        "type": "array",
                        "items": {
                            "type": "object",
                            "additionalProperties": {"type": "integer"},
                        },
                    },
                    "id": {"title": "Id", "description": "id", "type": "integer"},
                    "title": {
                        "title": "Title",
                        "description": "title",
                        "maxLength": 30,
                        "type": "string",
                    },
                },
                "required": ["title"],
            }
        },
    }

    class ThreadWithMessageListSchema(ModelSchema):
        messages: List[MessageSchema]

        class Config:
            model = Thread

    assert ThreadWithMessageListSchema.schema() == {
        "title": "ThreadWithMessageListSchema",
        "description": "A thread of messages.",
        "type": "object",
        "properties": {
            "messages": {
                "title": "Messages",
                "type": "array",
                "items": {"$ref": "#/definitions/MessageSchema"},
            },
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "title": {
                "title": "Title",
                "description": "title",
                "maxLength": 30,
                "type": "string",
            },
        },
        "required": ["messages", "title"],
        "definitions": {
            "MessageSchema": {
                "title": "MessageSchema",
                "description": "A message posted in a thread.",
                "type": "object",
                "properties": {
                    "id": {"title": "Id", "description": "id", "type": "integer"},
                    "content": {
                        "title": "Content",
                        "description": "content",
                        "type": "string",
                    },
                    "created_at": {
                        "title": "Created At",
                        "description": "created_at",
                        "type": "string",
                        "format": "date-time",
                    },
                    "thread": {
                        "title": "Thread",
                        "description": "id",
                        "type": "integer",
                    },
                },
                "required": ["content", "created_at", "thread"],
            }
        },
    }


@pytest.mark.django_db
def test_one_to_one():
    """
    Test forward one-to-one relationships.
    """

    class UserSchema(ModelSchema):
        class Config:
            model = User

    class ProfileSchema(ModelSchema):
        class Config:
            model = Profile

    assert ProfileSchema.schema() == {
        "title": "ProfileSchema",
        "description": "A user's profile.",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "user": {"title": "User", "description": "id", "type": "integer"},
            "website": {
                "title": "Website",
                "description": "website",
                "default": "",
                "maxLength": 200,
                "type": "string",
            },
            "location": {
                "title": "Location",
                "description": "location",
                "default": "",
                "maxLength": 100,
                "type": "string",
            },
        },
        "required": ["user"],
    }

    class ProfileWithUserSchema(ModelSchema):
        user: UserSchema

        class Config:
            model = Profile

    assert ProfileWithUserSchema.schema() == {
        "title": "ProfileWithUserSchema",
        "description": "A user's profile.",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "user": {"$ref": "#/definitions/UserSchema"},
            "website": {
                "title": "Website",
                "description": "website",
                "default": "",
                "maxLength": 200,
                "type": "string",
            },
            "location": {
                "title": "Location",
                "description": "location",
                "default": "",
                "maxLength": 100,
                "type": "string",
            },
        },
        "required": ["user"],
        "definitions": {
            "UserSchema": {
                "title": "UserSchema",
                "description": "A user of the application.",
                "type": "object",
                "properties": {
                    "profile": {
                        "title": "Profile",
                        "description": "id",
                        "type": "integer",
                    },
                    "id": {"title": "Id", "description": "id", "type": "integer"},
                    "first_name": {
                        "title": "First Name",
                        "description": "first_name",
                        "maxLength": 50,
                        "type": "string",
                    },
                    "last_name": {
                        "title": "Last Name",
                        "description": "last_name",
                        "maxLength": 50,
                        "type": "string",
                    },
                    "email": {
                        "title": "Email",
                        "description": "email",
                        "maxLength": 254,
                        "type": "string",
                    },
                    "created_at": {
                        "title": "Created At",
                        "description": "created_at",
                        "type": "string",
                        "format": "date-time",
                    },
                    "updated_at": {
                        "title": "Updated At",
                        "description": "updated_at",
                        "type": "string",
                        "format": "date-time",
                    },
                },
                "required": ["first_name", "email", "created_at", "updated_at"],
            }
        },
    }


@pytest.mark.django_db
def test_one_to_one_reverse():
    """
    Test reverse one-to-one relationships.
    """

    class ProfileSchema(ModelSchema):
        class Config:
            model = Profile

    class UserSchema(ModelSchema):
        class Config:
            model = User

    assert ProfileSchema.schema() == {
        "title": "ProfileSchema",
        "description": "A user's profile.",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "user": {"title": "User", "description": "id", "type": "integer"},
            "website": {
                "title": "Website",
                "description": "website",
                "default": "",
                "maxLength": 200,
                "type": "string",
            },
            "location": {
                "title": "Location",
                "description": "location",
                "default": "",
                "maxLength": 100,
                "type": "string",
            },
        },
        "required": ["user"],
    }

    class UserWithProfileSchema(ModelSchema):
        profile: ProfileSchema

        class Config:
            model = User

    assert UserWithProfileSchema.schema() == {
        "title": "UserWithProfileSchema",
        "description": "A user of the application.",
        "type": "object",
        "properties": {
            "profile": {"$ref": "#/definitions/ProfileSchema"},
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "first_name": {
                "title": "First Name",
                "description": "first_name",
                "maxLength": 50,
                "type": "string",
            },
            "last_name": {
                "title": "Last Name",
                "description": "last_name",
                "maxLength": 50,
                "type": "string",
            },
            "email": {
                "title": "Email",
                "description": "email",
                "maxLength": 254,
                "type": "string",
            },
            "created_at": {
                "title": "Created At",
                "description": "created_at",
                "type": "string",
                "format": "date-time",
            },
            "updated_at": {
                "title": "Updated At",
                "description": "updated_at",
                "type": "string",
                "format": "date-time",
            },
        },
        "required": ["profile", "first_name", "email", "created_at", "updated_at"],
        "definitions": {
            "ProfileSchema": {
                "title": "ProfileSchema",
                "description": "A user's profile.",
                "type": "object",
                "properties": {
                    "id": {"title": "Id", "description": "id", "type": "integer"},
                    "user": {"title": "User", "description": "id", "type": "integer"},
                    "website": {
                        "title": "Website",
                        "description": "website",
                        "default": "",
                        "maxLength": 200,
                        "type": "string",
                    },
                    "location": {
                        "title": "Location",
                        "description": "location",
                        "default": "",
                        "maxLength": 100,
                        "type": "string",
                    },
                },
                "required": ["user"],
            }
        },
    }


@pytest.mark.django_db
def test_generic_relation():
    """
    Test generic foreign-key relationships.
    """

    class TaggedSchema(ModelSchema):
        class Config:
            model = Tagged

    assert TaggedSchema.schema() == {
        "title": "TaggedSchema",
        "description": "Tagged(id, slug, content_type, object_id)",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "slug": {
                "title": "Slug",
                "description": "slug",
                "maxLength": 50,
                "type": "string",
            },
            "content_type": {
                "title": "Content Type",
                "description": "id",
                "type": "integer",
            },
            "object_id": {
                "title": "Object Id",
                "description": "object_id",
                "type": "integer",
            },
            "content_object": {
                "title": "Content Object",
                "description": "content_object",
                "type": "integer",
            },
        },
        "required": ["slug", "content_type", "object_id", "content_object"],
    }

    class BookmarkSchema(ModelSchema):
        # FIXME: I added this because for some reason in 2.2 the GenericRelation field
        # ends up required, but in 3 it does not.
        tags: List[Dict[str, int]] = None

        class Config:
            model = Bookmark

    assert BookmarkSchema.schema() == {
        "title": "BookmarkSchema",
        "description": "Bookmark(id, url)",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "url": {
                "title": "Url",
                "description": "url",
                "maxLength": 200,
                "type": "string",
            },
            "tags": {
                "title": "Tags",
                "type": "array",
                "items": {
                    "type": "object",
                    "additionalProperties": {"type": "integer"},
                },
            },
        },
        "required": ["url"],
    }

    class BookmarkWithTaggedSchema(ModelSchema):

        tags: List[TaggedSchema]

        class Config:
            model = Bookmark

    assert BookmarkWithTaggedSchema.schema() == {
        "title": "BookmarkWithTaggedSchema",
        "description": "Bookmark(id, url)",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "url": {
                "title": "Url",
                "description": "url",
                "maxLength": 200,
                "type": "string",
            },
            "tags": {
                "title": "Tags",
                "type": "array",
                "items": {"$ref": "#/definitions/TaggedSchema"},
            },
        },
        "required": ["url", "tags"],
        "definitions": {
            "TaggedSchema": {
                "title": "TaggedSchema",
                "description": "Tagged(id, slug, content_type, object_id)",
                "type": "object",
                "properties": {
                    "id": {"title": "Id", "description": "id", "type": "integer"},
                    "slug": {
                        "title": "Slug",
                        "description": "slug",
                        "maxLength": 50,
                        "type": "string",
                    },
                    "content_type": {
                        "title": "Content Type",
                        "description": "id",
                        "type": "integer",
                    },
                    "object_id": {
                        "title": "Object Id",
                        "description": "object_id",
                        "type": "integer",
                    },
                    "content_object": {
                        "title": "Content Object",
                        "description": "content_object",
                        "type": "integer",
                    },
                },
                "required": ["slug", "content_type", "object_id", "content_object"],
            }
        },
    }

    class ItemSchema(ModelSchema):

        tags: List[TaggedSchema]

        class Config:
            model = Item

    # Test without defining a GenericRelation on the model
    assert ItemSchema.schema() == {
        "title": "ItemSchema",
        "description": "Item(id, name, item_list)",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "name": {
                "title": "Name",
                "description": "name",
                "maxLength": 100,
                "type": "string",
            },
            "item_list": {
                "title": "Item List",
                "description": "id",
                "type": "integer",
            },
            "tags": {
                "title": "Tags",
                "type": "array",
                "items": {"$ref": "#/definitions/TaggedSchema"},
            },
        },
        "required": ["name", "item_list", "tags"],
        "definitions": {
            "TaggedSchema": {
                "title": "TaggedSchema",
                "description": "Tagged(id, slug, content_type, object_id)",
                "type": "object",
                "properties": {
                    "id": {"title": "Id", "description": "id", "type": "integer"},
                    "slug": {
                        "title": "Slug",
                        "description": "slug",
                        "maxLength": 50,
                        "type": "string",
                    },
                    "content_type": {
                        "title": "Content Type",
                        "description": "id",
                        "type": "integer",
                    },
                    "object_id": {
                        "title": "Object Id",
                        "description": "object_id",
                        "type": "integer",
                    },
                    "content_object": {
                        "title": "Content Object",
                        "description": "content_object",
                        "type": "integer",
                    },
                },
                "required": ["slug", "content_type", "object_id", "content_object"],
            }
        },
    }


@pytest.mark.django_db
def test_m2m_reverse():
    class ExpertSchema(ModelSchema):
        class Config:
            model = Expert

    class CaseSchema(ModelSchema):
        class Config:
            model = Case

    assert ExpertSchema.schema() == {
        "title": "ExpertSchema",
        "description": "Expert(id, name)",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "name": {
                "title": "Name",
                "description": "name",
                "maxLength": 128,
                "type": "string",
            },
            "cases": {
                "title": "Cases",
                "description": "id",
                "type": "array",
                "items": {
                    "type": "object",
                    "additionalProperties": {"type": "integer"},
                },
            },
        },
        "required": ["name", "cases"],
    }

    assert CaseSchema.schema() == {
        "title": "CaseSchema",
        "description": "Case(id, name, details)",
        "type": "object",
        "properties": {
            "related_experts": {
                "title": "Related Experts",
                "description": "id",
                "type": "array",
                "items": {
                    "type": "object",
                    "additionalProperties": {"type": "integer"},
                },
            },
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "name": {
                "title": "Name",
                "description": "name",
                "maxLength": 128,
                "type": "string",
            },
            "details": {"title": "Details", "description": "details", "type": "string"},
        },
        "required": ["name", "details"],
    }
    case = Case.objects.create(name="My Case", details="Some text data.")
    expert = Expert.objects.create(name="My Expert")
    case_schema = CaseSchema.from_django(case)
    expert_schema = ExpertSchema.from_django(expert)
    assert case_schema.dict() == {
        "related_experts": [],
        "id": 1,
        "name": "My Case",
        "details": "Some text data.",
    }
    assert expert_schema.dict() == {"id": 1, "name": "My Expert", "cases": []}

    expert.cases.add(case)
    case_schema = CaseSchema.from_django(case)
    expert_schema = ExpertSchema.from_django(expert)
    assert case_schema.dict() == {
        "related_experts": [{"id": 1}],
        "id": 1,
        "name": "My Case",
        "details": "Some text data.",
    }
    assert expert_schema.dict() == {"id": 1, "name": "My Expert", "cases": [{"id": 1}]}

    class CustomExpertSchema(ModelSchema):
        """Custom schema"""

        name: Optional[str]

        class Config:
            model = Expert

    class CaseSchema(ModelSchema):
        related_experts: List[CustomExpertSchema]

        class Config:
            model = Case

    assert CaseSchema.schema() == {
        "title": "CaseSchema",
        "description": "Case(id, name, details)",
        "type": "object",
        "properties": {
            "related_experts": {
                "title": "Related Experts",
                "type": "array",
                "items": {"$ref": "#/definitions/CustomExpertSchema"},
            },
            "id": {"title": "Id", "description": "id", "type": "integer"},
            "name": {
                "title": "Name",
                "description": "name",
                "maxLength": 128,
                "type": "string",
            },
            "details": {"title": "Details", "description": "details", "type": "string"},
        },
        "required": ["related_experts", "name", "details"],
        "definitions": {
            "CustomExpertSchema": {
                "title": "CustomExpertSchema",
                "description": "Custom schema",
                "type": "object",
                "properties": {
                    "id": {"title": "Id", "description": "id", "type": "integer"},
                    "name": {"title": "Name", "type": "string"},
                    "cases": {
                        "title": "Cases",
                        "description": "id",
                        "type": "array",
                        "items": {
                            "type": "object",
                            "additionalProperties": {"type": "integer"},
                        },
                    },
                },
                "required": ["cases"],
            }
        },
    }

    case_schema = CaseSchema.from_django(case)
    assert case_schema.dict() == {
        "related_experts": [{"id": 1, "name": "My Expert", "cases": [{'id': 1}]}],
        "id": 1,
        "name": "My Case",
        "details": "Some text data.",
    }


@pytest.mark.django_db
def test_alias():
    class ProfileSchema(ModelSchema):
        first_name: str = Field(alias='user__first_name')

        class Config:
            model = Profile

    assert ProfileSchema.schema() == {
        'title': 'ProfileSchema',
        'description': "A user's profile.",
        'type': 'object',
        'properties': {
            'id': {
                'title': 'Id',
                'description': 'id',
                'type': 'integer'
            },
            'user': {
                'title': 'User',
                'description': 'id',
                'type': 'integer'
            },
            'website': {
                'title': 'Website',
                'description': 'website',
                'default': '', 'maxLength': 200,
                'type': 'string'
            },
            'location': {
                'title': 'Location',
                'description': 'location',
                'default': '',
                'maxLength': 100,
                'type': 'string'
            },
            'user__first_name': {
                'title': 'User  First Name',
                'type': 'string'
            }
        },
        'required': ['user', 'user__first_name']
    }

    user = User.objects.create(first_name="Jack")
    profile = Profile.objects.create(
        user=user, website='www.github.com', location='Europe')
    assert ProfileSchema.from_django(profile).dict() == {'first_name': 'Jack',
                                                         'id': 1,
                                                         'location': 'Europe',
                                                         'user': 1,
                                                         'website': 'www.github.com'}



from setuptools import find_packages, setup


__version__ = "0.7.0"


def get_long_description():
    return open("README.md", "r", encoding="utf8").read()


setup(
    name="djantic",
    version=__version__,
    packages=find_packages(),
    license="MIT",
    url="https://github.com/jordaneremieff/djantic/",
    description="Pydantic model support for Django ORM",
    long_description=get_long_description(),
    python_requires=">=3.7",
    package_data={"djantic": ["py.typed"]},
    long_description_content_type="text/markdown",
    author="Jordan Eremieff",
    author_email="jordan@eremieff.com",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Framework :: Django",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)
